import tkinter as tk
from tkinter import ttk
import copy

from tktooltip import ToolTip
from segdan.utils.constants import ClusteringModelName
from segdan.utils.confighandler import ConfigHandler

class ClusteringModelForm():
    def __init__(self, parent, allow_grid=False, models=ConfigHandler.CONFIGURATION_VALUES["clustering_models"], config_data={}, edit=False):
            self.top = tk.Toplevel(parent)
            self.allow_grid = allow_grid

            self.top.title("Clustering model selection")

            self.top.geometry("450x275")
            self.top.resizable(False, False)

            self.model_selected = tk.StringVar(value="")
            self.grid_search = tk.BooleanVar(value=False)

            self.config_data = config_data
            self.temp_config_data = {}

            if edit and config_data:
                self.model_selected.set(list(config_data.keys())[0])
                self.model_name = self.model_selected.get()
            else:
                self.model_name = ""
            self.edit = edit
            self.models = models
            self.linkages = ConfigHandler.CONFIGURATION_VALUES["linkages"]
            self.default_hyperparameters = ConfigHandler.DEFAULT_CLUSTERING_HYPERPARAMETERS

            self.top.grid_columnconfigure(0, weight=1)
            self.top.grid_columnconfigure(1, weight=1)
            self.top.grid_columnconfigure(2, weight=1)
            self.top.grid_columnconfigure(3, weight=1)

            self.model_frame = ttk.LabelFrame(self.top, text="Hyperparameters", padding=(20, 10))
            self.model_frame.grid(row=2, column=0, columnspan=4, padx=10, pady=10, sticky="ew")
            
            self.create_controls()
            
            self.on_model_change(None)
            
            if edit and config_data:
                self.model_selected.set(list(config_data.keys())[0])
                self.model_name = self.model_selected.get()
                
                model_data = config_data.get(self.model_name, {})
                grid_params = any("_range" in key for key in model_data.keys())  

                if grid_params:
                    self.grid_search.set(True) 
                    self.on_grid_search_toggle()


            self.top.protocol("WM_DELETE_WINDOW", self.clear_and_close_form)
            

    def create_controls(self):
        for i in range(6):
            self.top.grid_columnconfigure(i, weight=1)

        center_frame = tk.Frame(self.top)
        center_frame.grid(row=1, column=0, columnspan=3, pady=10) 

        clustering_model_label = tk.Label(center_frame, text="Clustering model *")
        clustering_model_label.grid(row=1, column=0, padx=10, sticky="e")
        ToolTip(clustering_model_label, msg="Choose a clustering algorithm from the dropdown list.")

        model_dropdown = ttk.Combobox(center_frame, textvariable=self.model_selected, values=self.models, state="readonly", width=15)
        model_dropdown.grid(row=1, column=1, padx=10, pady=5, sticky="w") 
        model_dropdown.bind("<<ComboboxSelected>>", self.on_model_change)

        if self.allow_grid:
            grid_search_label = tk.Label(center_frame, text="Use grid search")
            grid_search_label.grid(row=1, column=2, sticky="w", padx=10)
            ToolTip(grid_search_label, msg="Check this box to perform a grid search for hyperparameters.")

            grid_search_checkbox = tk.Checkbutton(center_frame, variable=self.grid_search, command=self.on_grid_search_toggle)
            grid_search_checkbox.grid(row=1, column=3, padx=10, pady=5, sticky="w")

        btn_frame = tk.Frame(self.top)
        btn_frame.grid(row=3, column=0, columnspan=4, pady=20)  

        if self.edit:
            tk.Button(btn_frame, text="Delete model", command=self.delete_model).grid(row=0, column=0, padx=5)
        tk.Button(btn_frame, text="Save", command=self.validate_and_close_form).grid(row=0, column=1, padx=5)
        tk.Button(btn_frame, text="Cancel", command=self.clear_and_close_form).grid(row=0, column=2, padx=5)
        

    def create_model_frame(self):
        self.model_frame = ttk.LabelFrame(self.top, text="Hyperparameters", padx=10, pady=10, padding=(20, 10))
        self.model_frame.grid(row=2, column=0, columnspan=3, padx=10, pady=10, sticky="ew")

    def clear_model_frame(self):
        for widget in self.model_frame.winfo_children():
            widget.destroy()

    def clear_form(self):
        for widget in self.model_frame.grid_slaves():
            if int(widget.grid_info()["row"]) > 1:  
                widget.grid_forget()


    def on_model_change(self, event):
        
        model = self.model_selected.get().lower() 
        self.clear_form()
        self.clear_model_frame()
        
        if model not in self.config_data:
            self.temp_config_data = {}  
        else:
            self.temp_config_data = self.config_data.copy()

        if model != "best model":
            self.model_frame.grid()

        if model == ClusteringModelName.KMEANS.value:
             self.show_kmeans_params()

        elif model == ClusteringModelName.AGGLOMERATIVE.value:
            self.show_agglomerative_params()

        elif model == ClusteringModelName.DBSCAN.value:
            self.show_dbscan_params()

        elif model == ClusteringModelName.OPTICS.value:
            self.show_optics_params()     
        elif model == "best model":
            self.model_frame.grid_remove()
    
    def on_grid_search_toggle(self):
        model = self.model_selected.get().lower()

        if self.grid_search.get():
            if model not in self.temp_config_data:
                self.temp_config_data[model] = {}

                grid_defaults = {
                    "eps": {"min": 0.1, "max": 1.0, "step": 0.1},
                    "min_samples": {"min": 2, "max": 10, "step": 1},
                    "num_clusters": {"min": 2, "max": 10, "step": 1},
                }
                
                for param_key, default_value in grid_defaults.items():
                    if param_key not in self.temp_config_data[model]:
                        self.temp_config_data[model][f"{param_key}_range"] = default_value
        else:
            for param_key in ["n_clusters","eps","min_samples"]:
                self.temp_config_data[model].pop(f"{param_key}_range", None)

        self.on_model_change(None)  

    def show_kmeans_params(self):
        self.create_param_field("kmeans", "Number of clusters", "n_clusters", row=0)
        self.create_param_field("kmeans", "Random state", "random_state", row=1)

    def show_agglomerative_params(self):
        self.create_param_field("agglomerative", "Number of clusters", "n_clusters", row=0)
        self.create_categorical_param_field("agglomerative", "Linkage", "linkage", row=1, options=self.linkages)

    def show_dbscan_params(self):
        self.create_param_field("dbscan", "Eps ", "eps", row=0)
        self.create_param_field("dbscan", "Min samples", "min_samples", row=1)

    def show_optics_params(self):
        self.create_param_field("optics", "Min samples ", "min_samples", row=0)

    def update_int_grid_value(self, model, param_key, range_key, value):
        try:
            num_value = int(value) if value.isdigit() else float(value)
        except ValueError:
            num_value = None

        if model not in self.temp_config_data:
            self.temp_config_data[model] = {}

        if range_key:
            self.temp_config_data[model].setdefault(param_key, {})[range_key] = num_value
        else:
            self.temp_config_data[model][param_key] = num_value

    def update_range_value(self, model, param_key, range_key, value):
        try:
            num_value = int(value) if value else None
        except ValueError:
            num_value = float(value) if value else None

        if model not in self.temp_config_data:
            self.temp_config_data[model] = {}

        if range_key:
            self.temp_config_data[model].setdefault(f"{param_key}", {})[range_key] = num_value
        else:
            self.temp_config_data[model][param_key] = num_value

    def update_single_value(self, model, param_key, value):
        try:
            num_value = int(value) if value else None
        except ValueError:
            num_value = float(value) if value else None

        if model not in self.temp_config_data:
            self.temp_config_data[model] = {}

        self.temp_config_data[model][param_key] = num_value


    def create_param_field(self, model, label_text, param_key, row):

        if model not in self.temp_config_data:
            self.temp_config_data[model] = {}

        if any(key.endswith("_range") for key in self.temp_config_data[model]) and not self.grid_search.get():
            self.grid_search.set(True)


        param_label = tk.Label(self.model_frame, text=label_text)
        param_label.grid(row=row, column=0, padx=5, pady=5, sticky="w")

        if param_key == "random_state":
            ToolTip(param_label, msg="Seed for random number generation, ensuring reproducibility.")
        elif param_key == "n_clusters":
            ToolTip(param_label, msg="Number of clusters to form; must be at least 2.")
        elif param_key == "min_samples":
            ToolTip(param_label, msg="Minimum points required to form a dense region; must be at least 2.")
        elif param_key == "eps":
            ToolTip(param_label, msg="Maximum distance between points to be considered neighbors.")


        if param_key=="random_state" or not self.grid_search.get():
            value = self.temp_config_data[model].get(param_key, self.default_hyperparameters.get(param_key, 5))

            self.temp_config_data[model][param_key] = value

            entry = tk.Entry(self.model_frame, width=10)
            entry.grid(row=row, column=1, padx=5, pady=5)
            entry.insert(0, str(value))
            entry.bind("<KeyRelease>", lambda event: self.update_single_value(model, param_key, entry.get()))
            return
            
        param_key = f"{param_key}_range"
        value_range = self.temp_config_data[model].get(param_key, copy.deepcopy(self.default_hyperparameters.get(param_key, {"min": 2, "max": 10, "step": 1})))

        self.temp_config_data[model].setdefault(param_key, value_range)

        labels_entries = [
        ("Min:", "min", 1, 2),
        ("Max:", "max", 3, 4),
        ("Step:", "step", 5, 6),
        ]

        for text, key, col_label, col_entry in labels_entries:
            tk.Label(self.model_frame, text=text).grid(row=row, column=col_label, padx=2, pady=5)
            entry = tk.Entry(self.model_frame, width=5)
            entry.grid(row=row, column=col_entry, padx=2, pady=5)
            entry.insert(0, str(value_range[key]))
            entry.bind("<KeyRelease>", lambda event, k=key, e=entry: self.update_range_value(model, param_key, k, e.get()))

    def update_selected_values(self, event, listbox, options, model, param_key):
        selected_indices = listbox.curselection()  
        selected_values = [options[i] for i in selected_indices]  
        self.temp_config_data[model][param_key] = selected_values  

    def update_selected_values_listbox(self, event, listbox, options, model, param_key):
        selected_indices = listbox.curselection()
        self.temp_config_data[model][param_key] = [options[i] for i in selected_indices]

    def update_selected_value_combobox(self, event, combobox, model, param_key):
        value = combobox.get()
        self.temp_config_data[model][param_key] = value
        
    def create_categorical_param_field(self, model, label_text, param_key, row, options):
        param_label = tk.Label(self.model_frame, text=label_text)
        param_label.grid(row=row, column=0, padx=5, pady=5, sticky="w")

        ToolTip(param_label, msg="Select the linkage criterion for cluster merging.")

        if param_key not in self.temp_config_data[model]:
            self.temp_config_data[model][param_key] = options[0]

        value = self.temp_config_data[model][param_key]

        if self.allow_grid and self.grid_search.get():
            param_key = "linkages"
            
            self.temp_config_data[model].setdefault(param_key, [])
            
            listbox = tk.Listbox(self.model_frame, selectmode="multiple", height=len(options), exportselection=0)
            
            for option in options:
                listbox.insert(tk.END, option)
            
            listbox.grid(row=row, column=1, columnspan=2, padx=5, pady=5)

            selected_values = set(self.temp_config_data[model][param_key])  
        
            for i, option in enumerate(options):
                if option in selected_values:
                    listbox.selection_set(i)

            listbox.bind('<<ListboxSelect>>', lambda event: self.update_selected_values_listbox(event, listbox, options, model, param_key))
            return
        
        param_key = "linkage"
        
        combobox = ttk.Combobox(self.model_frame, textvariable=self.temp_config_data[model][param_key], values=options, state="readonly", width=10)
        combobox.grid(row=row, column=1, padx=5, pady=5)
        combobox.set(value)
        combobox.bind("<<ComboboxSelected>>", lambda event: self.update_selected_value_combobox(event, combobox, model, param_key))

    def validate_numeric(self, value):
        return value.isdigit() or value == "" 

    def validate_numeric_and_float(self, value):
        if value == "" or value.isdigit():
            return True
        if value.count('.') == 1 and value.replace('.', '').isdigit():
            return True
        return False
    
    def validate_form(self):

        model = self.model_selected.get().lower()

        if model == "":
            tk.messagebox.showerror("Error", "No model selected. Please select a clustering model.")
            return False

        if model in self.temp_config_data:
            params = self.temp_config_data[model]
            

            if model in ["kmeans", "agglomerative"]:
                if "n_clusters" in params:
                    n_clusters = params["n_clusters"]
                    if n_clusters == "":
                        tk.messagebox.showerror("Error", f"Number of clusters must be specified.")
                        return False
                    if n_clusters < 2:
                        tk.messagebox.showerror("Error", f"Number of clusters with {model} model must be at least 2.")
                        return False
                if "linkages" in params:
                    linkages = params["linkages"]
                    if len(linkages) == 0:
                        tk.messagebox.showerror("Error", "At least one linkage for agglomerative clustering must be selected.")
                        return False
                if "n_clusters_range" in params:
                    n_clusters_min = params["n_clusters_range"]["min"]
                    n_clusters_max = params["n_clusters_range"]["max"]
                    if n_clusters_min == "" or n_clusters_max == "" or params["n_clusters_range"]["step"] == "":
                        tk.messagebox.showerror("Error", "Cluster range values cannot be empty.")
                        return False
                    
                    if n_clusters_min < 2:
                        tk.messagebox.showerror("Error", f"Minimum number of clusters with {model} model must be at least 2.")
                        return False

                    if n_clusters_min > n_clusters_max:
                        tk.messagebox.showerror("Error", f"Min value of clusters range must be less than max value.")
                        return False
                
            if model == ClusteringModelName.DBSCAN.value:
                if "eps_range" in params:
                    min_eps_range = params["eps_range"]["min"]
                    max_eps_range = params["eps_range"]["max"]
                    
                    if min_eps_range == "" or max_eps_range == "" or params["eps_range"]["step"] == "":
                        tk.messagebox.showerror("Error", "Eps range values cannot be empty.")
                        return False

                    if min_eps_range > max_eps_range:
                        tk.messagebox.showerror("Error", f"Min value of eps range must be less than max value.")
                        return False

            if model in [ClusteringModelName.DBSCAN.value,ClusteringModelName.OPTICS.value]:
                if "min_samples" in params:
                    min_samples = params["min_samples"]
                    if min_samples == "":
                        tk.messagebox.showerror("Error", "Min samples value cannot be empty.")
                        return False
            
                    if min_samples < 2:
                        tk.messagebox.showerror("Error", f"Min samples with {model} model must be at least 2.")
                        return False
                elif "min_samples_range" in params:
                    min_samples_min = params["min_samples_range"]["min"]
                    min_samples_max = params["min_samples_range"]["max"]
                    if min_samples_min == "" or min_samples_max == "" or params["min_samples_range"]["step"] == "":
                        tk.messagebox.showerror("Error", "Min samples range values cannot be empty.")
                        return False

                    if min_samples_min < 2:
                        tk.messagebox.showerror("Error", f"Minimum min samples with {model} model must be at least 2.")
                        return False
                    if min_samples_min > min_samples_max:
                        tk.messagebox.showerror("Error", f"Min value of min samples range must be less than max value.")
                        return False

        return True

    def clear_and_close_form(self):
        self.temp_config_data = {}
        self.grid_search.set(False)
        self.top.destroy()

    def update_params(self, model):
        grid_search_removals = {
            "kmeans": ["n_clusters"],
            "agglomerative": ["linkage", "n_clusters"],
            "dbscan": ["eps", "min_samples"],
            "optics": ["min_samples"]
        }

        normal_removals = {
            "kmeans": ["n_clusters_range"],
            "agglomerative": ["linkages", "n_clusters_range"],
            "dbscan": ["eps_range", "min_samples_range"],
            "optics": ["min_samples_range"]
        }

        if self.grid_search.get():
            if model in grid_search_removals:
                for key in grid_search_removals[model]:
                    self.temp_config_data[model].pop(key, None)
            return

        if model in normal_removals:
            for key in normal_removals[model]:
                self.temp_config_data[model].pop(key, None)


    def validate_and_close_form(self):
        if not self.validate_form():
            return

        model = self.model_selected.get().lower()
        
        self.update_params(model)

        self.config_data[model] = self.temp_config_data[model]

        self.top.destroy() 
           

    def delete_model(self):
        if self.model_name == "":
            tk.messagebox.showerror("Error", "You must select a model to delete.")
            return
        
        delete = tk.messagebox.askyesno("Delete model", "Are you sure you want to delete this model and all of its information?")

        if not delete:
            return
    
        deleted_model = self.model_selected.get().lower()
        self.config_data.pop(deleted_model, None)
        self.temp_config_data = {}

        self.top.destroy()
        tk.messagebox.showinfo("Delete model", f"Model {deleted_model} has been successfully deleted from the configuration.")




from ultralytics.data.converter import convert_segment_masks_to_yolo_seg
from segdan.converters.converter import Converter

class MultilabelToYOLOConverter(Converter):

    def __init__(self, input_data: str, output_dir: str, num_classes: int):
        super().__init__(input_data, output_dir)
        self.num_classes = num_classes

    def convert(self):
        convert_segment_masks_to_yolo_seg(masks_dir=self.input_data, output_dir=self.output_dir, classes=self.num_classes)



import shutil
import tkinter as tk
from tkinter import ttk, filedialog, messagebox
from tktooltip import ToolTip

from segdan.utils.constants import LabelFormat
from segdan.forms.colorconfigform import ColorConfigForm
from segdan.forms.formutils import FormUtils
from segdan.utils.confighandler import ConfigHandler

import os

class GeneralConfigFrame(ttk.Frame):
    def __init__(self, parent, controller, config_data, final_dict):
        ttk.Frame.__init__(self, parent)
        
        self.config_data = config_data
        self.controller = controller
        self.top = parent

        if "general_data" not in self.config_data:
            self.config_data["general_data"] = {
                "class_map_file" : self.config_data.get("class_map_file", tk.StringVar(value="")),
                "label_format": self.config_data.get("label_format", tk.StringVar(value="txt")),
                "color_dict": self.config_data.get("color_dict", {}),
                "class_mapping": self.config_data.get("class_mapping", {}),
                "image_path": self.config_data.get("image_path", tk.StringVar(value="")),
                "label_path": self.config_data.get("label_path", tk.StringVar(value="")),
                "output_path": self.config_data.get("output_path", tk.StringVar(value="")),
                "verbose": self.config_data.get("verbose", tk.BooleanVar(value=False)),
                "binary": self.config_data.get("binary", tk.BooleanVar(value=False)),
                "background": self.config_data.get("background", tk.StringVar(value="0")),
                "threshold": self.config_data.get("threshold", tk.StringVar(value="255"))
            }
            

        label_title = ttk.Label(self, text="General configuration", font=("Arial", 18, "bold"))
        label_title.grid(row=0, column=0, columnspan=5, pady=(20,10), padx=10)

        self.general_data = self.config_data["general_data"]

        self.general_config_widgets()
        self.create_widgets()
        
        button_frame = ttk.Frame(self.general_frame)
        button_frame.grid(row=15, column=0, columnspan=5, pady=10, sticky="e")  

        self.general_frame.grid_rowconfigure(7, weight=0)
        self.general_frame.grid_columnconfigure(0, weight=1)

        button_back = ttk.Button(button_frame, text="Back", command=lambda: controller.show_frame("IntroductionFrame"))
        button_back.grid(row=0, column=0, padx=50, pady=5, sticky="w")

        button_next = ttk.Button(button_frame, text="Next", command=self.next)
        button_next.grid(row=0, column=1, pady=5, sticky="e")

        button_frame.grid_columnconfigure(0, weight=0)
        button_frame.grid_columnconfigure(1, weight=0)

        self.on_format_change()
        self.toggle_load_button()

    def validate_numeric(self, value):
        if value == "" or value.isdigit():
            return True
        return False

    def validate_threshold(self, value):
        if value == "" or (value.isdigit() and 0 <= int(value) <= 255):
            return True
        return False

    def create_widgets(self):

        vcmd = (self.top.register(self.validate_numeric), "%P")
        val_threshold = (self.top.register(self.validate_threshold), "%P")

        self.color_map_label = tk.Label(self.general_labelframe, text="Color mapping")
        self.transformation_color_dict_text = tk.Text(self.general_labelframe, height=5, width=20, state="disabled")
        ToolTip(self.color_map_label, msg="Mapping of class IDs to colors.")

        self.transformation_color_dict_text.insert("1.0", str(self.general_data['color_dict']))  
        self.transformation_color_dict_text.config(state="disabled")  

        self.add_color_bt = tk.Button(self.general_labelframe, text="Add dictionary", command=lambda: self.open_color_form())
        ToolTip(self.add_color_bt, msg="Add colors to the dictionary.")

        self.class_map_label = tk.Label(self.general_labelframe, text="Class mapping")
        self.class_mapping = ttk.Entry(self.general_labelframe, textvariable=self.general_data["class_map_file"], width=50, state="readonly")
        self.class_map_bt = tk.Button(self.general_labelframe, text="Select file 📄", command= lambda: self.select_file("class_map_file", "txt"))
        ToolTip(self.class_map_label, msg="Mapping of class IDs to names (format: ID:Name)")
        ToolTip(self.class_map_bt, msg="Select a file containing the class mapping.")

        self.clear_classmap_bt = tk.Button(self.general_labelframe, text="Delete file 📄", command= lambda: self.delete_file())
        ToolTip(self.clear_classmap_bt, msg="Clears class mapping configuration.")
        
        self.image_path_label = tk.Label(self.general_labelframe, text="Image path *")
        self.entry_images = ttk.Entry(self.general_labelframe, textvariable=self.general_data["image_path"], width=50, state="readonly")
        self.img_path_bt = tk.Button(self.general_labelframe, text="Select folder 📂", command=lambda: self.select_folder("image_path"))
        ToolTip(self.image_path_label, msg="Path to the folder containing the images of the dataset.")
        ToolTip(self.img_path_bt, msg="Select the folder containing the images.")
        
        self.label_path_label = tk.Label(self.general_labelframe, text="Label path *")
        self.label_entry = ttk.Entry(self.general_labelframe, textvariable=self.general_data["label_path"], width=50, state="readonly")
        self.label_path_bt = tk.Button(self.general_labelframe, text="Select folder 📂", command=lambda: self.select_folder("label_path"))
        ToolTip(self.label_path_label, msg="Path to the folder containing the labels of the dataset.")
        ToolTip(self.label_path_bt, msg="Select the folder containing the labels.")
        
        self.label_file_bt = tk.Button(self.general_labelframe, text="Select file 📄", command=lambda: self.select_file("label_path", "json"))
        ToolTip(self.label_file_bt, msg="Select the annotation file.")
       
        self.output_path_label = tk.Label(self.general_labelframe, text="Output path *")
        self.output_entry = ttk.Entry(self.general_labelframe, textvariable=self.general_data["output_path"], width=50, state="readonly")
        self.output_path_bt = tk.Button(self.general_labelframe, text="Select folder 📂", command=lambda: self.select_folder("output_path"))
        ToolTip(self.label_path_label, msg="Path to the folder for the results created by the program.")
        ToolTip(self.output_path_bt, msg="Select the folder where the results will be saved.")
        
        self.verbose_label = tk.Label(self.general_labelframe, text="Verbose")
        self.binary_label = tk.Label(self.general_labelframe, text="Binary")
        self.background_label = tk.Label(self.general_labelframe, text="Background")
        self.binary_threshold_label = tk.Label(self.general_labelframe, text="Binary threshold")
        
        self.verbose_checkbutton = ttk.Checkbutton(self.general_labelframe, variable=self.general_data["verbose"])
        ToolTip(self.verbose_label, msg="Enable detailed logging during execution.")
        
        self.binary_checkbutton = ttk.Checkbutton(self.general_labelframe, variable=self.general_data["binary"], command=lambda: FormUtils.toggle_label_entry(self.general_data["binary"], self.binary_threshold_label, 
                                                                                                                          self.threshold_entry, None, 12, 3))
        ToolTip(self.binary_label, msg="Set to True for binary segmentation (foreground/background).")
        
        self.background_entry = ttk.Entry(self.general_labelframe, textvariable=self.general_data["background"], width=5, validate="key", validatecommand=vcmd)
        ToolTip(self.background_label, msg="Class label assigned to background pixels.")

        self.threshold_entry = ttk.Entry(self.general_labelframe, textvariable=self.general_data["threshold"], width=5, validate="key", validatecommand=val_threshold)
        ToolTip(self.binary_threshold_label, msg="Binary label threshold value for two-class segmentation.\nIt must be between 0 and 255.")

        self.row+=1
        self.class_map_label.grid(row=self.row, column=0, padx=5, sticky="w")

        self.row+=1
        self.class_mapping.grid(row=self.row, column=0, columnspan=3, padx=5, pady=5)
        self.class_map_bt.grid(row=self.row, column=3, padx=5, pady=5)
        self.clear_classmap_bt.grid(row=self.row, column=6, padx=5, pady=5)
        self.clear_classmap_bt.grid_remove()

        self.row+=1
        self.color_map_label.grid(row=self.row, column=0, padx=5, pady=5, sticky="w")

        self.row+=1
        self.transformation_color_dict_text.grid(row=self.row, column=0, columnspan=3, pady=5, padx=5, sticky="we")
        self.add_color_bt.grid(row=self.row, column=3, padx=5, pady=5)        

        self.row+=1
        self.image_path_label.grid(row=self.row, column=0, padx=5, sticky="w")
        
        self.row+=1
        self.entry_images.grid(row=self.row, column=0, columnspan=3, padx=5, pady=5)
        self.img_path_bt.grid(row=self.row, column=3, padx=5, pady=5)

        self.row+=1
        self.label_path_label.grid(row=self.row, column=0, padx=5, sticky="w")

        self.row+=1
        self.label_entry.grid(row=self.row, column=0, columnspan=3, padx=5, pady=5)
        self.label_path_bt.grid(row=self.row, column=3, padx=5, pady=5)
        self.label_file_bt.grid(row=self.row, column=3, padx=5, pady=5)

        self.row+=1
        self.output_path_label.grid(row=self.row, column=0, padx=5, sticky="w")

        self.row+=1
        self.output_entry.grid(row=self.row, column=0, columnspan=3, padx=5, pady=5)
        self.output_path_bt.grid(row=self.row, column=3, padx=5, pady=5)

        self.row+=1
        self.verbose_label.grid(row=self.row, column=0, padx=5, pady=5)
        self.binary_label.grid(row=self.row, column=1, padx=5, pady=5)
        self.background_label.grid(row=self.row, column=2, padx=5, pady=5)
        self.binary_threshold_label.grid(row=self.row, column=3, padx=5, pady=5)

        self.row+=1
        self.verbose_checkbutton.grid(row=self.row, column=0, padx=5, pady=5)
        self.binary_checkbutton.grid(row=self.row, column=1, padx=5, pady=5)
        self.background_entry.grid(row=self.row, column=2, padx=5, pady=5)
        self.threshold_entry.grid(row=self.row, column=3, padx=5, pady=5)

        self.binary_threshold_label.grid_remove()
        self.threshold_entry.grid_remove()


    def open_color_form(self):
        color_form = ColorConfigForm(self.top, self.general_data["color_dict"])

        color_form.top.transient(self.top)
        color_form.top.grab_set()

        FormUtils.center_window(color_form, self.top)
        
        color_form.top.focus_set()

        self.top.wait_window(color_form.top)

        if color_form.colors:
            self.add_colors(color_form.colors)
            return
        
        self.transformation_color_dict_text.config(state="normal")  
        self.transformation_color_dict_text.delete(1.0, tk.END) 
        self.transformation_color_dict_text.config(state="disabled")

    def add_colors(self, colors):
        self.general_data["color_dict"] = colors

        colors_text = "\n".join([f"{color}: {class_id}" for color, class_id in colors.items()])

        self.transformation_color_dict_text.config(state="normal")  
        self.transformation_color_dict_text.delete(1.0, tk.END)  
        self.transformation_color_dict_text.insert(tk.END, colors_text)
        self.transformation_color_dict_text.config(state="disabled")

    def general_config_widgets(self):
        self.general_frame = tk.Frame(self, padx=10, pady=10)
        self.general_frame.grid(row=1, column=0, padx=10, pady=10)

        self.general_labelframe = ttk.LabelFrame(self.general_frame, text="General configuration", padding=(20,10))
        self.general_labelframe.grid(row=0, column=0, padx=5, pady=10, sticky="ew")

        self.grid_rowconfigure(0, weight=0)
        self.grid_columnconfigure(0, weight=1)

        self.row=0
        self.label_format_label = tk.Label(self.general_labelframe, text="Label format")
        self.label_format_label.grid(row=self.row, column=1, padx=10)

        self.row+=1
        self.label_format_dropdown = ttk.Combobox(self.general_labelframe, textvariable=self.general_data["label_format"], values=ConfigHandler.CONFIGURATION_VALUES["label_formats"], state="readonly", width=15)
        self.label_format_dropdown.grid(row=self.row, column=1, padx=10, pady=5) 
        ToolTip(self.label_format_label, msg="Select the label format.\nTXT for YOLO, JSON for COCO and mask for multiple labels per image.")

        self.general_data["label_format"].trace("w", lambda *args: self.on_format_change())
        self.general_data["class_map_file"].trace_add("write", lambda *args: self.toggle_load_button())

    def select_file(self, file_type, extension):
        filetypes = [("All Files", "*.*"), ("Text Files", "*.txt"), ("JSON Files", "*.json")]
        if extension == LabelFormat.TXT.value:
            filetypes = [filetypes[1]]
        elif extension == LabelFormat.JSON.value:
            filetypes = [filetypes[2]]
        file_path = filedialog.askopenfilename(title="Select File", filetypes=filetypes)
        
        if file_path:
            self.general_data[file_type].set(file_path) 
        
            if file_type == "class_map_file":
                self.load_class_map(file_path)
            
            self.toggle_load_button()

    def update_color_map(self):

        class_mapping = self.general_data["class_mapping"]
        color_map = self.general_data["color_dict"]

        for class_name, class_id in class_mapping.items():
            if class_id not in color_map:
                color_map[class_id] = ""

        self.general_data["color_dict"] = color_map
        self.add_colors(color_map)


    def load_class_map(self, file_path):
        try:
            with open(file_path) as f:
                data = f.readlines()
                class_mapping = {}
                for line in data:
                    parts = line.strip().split(":")
                    if len(parts) == 2:
                        class_id, class_name = parts
                        class_name = class_name.strip()
                        if class_name == "":
                            messagebox.showerror("Class file configuration error", f"ID {class_id} does not have any class name assigned.")
                            return
                        class_mapping[int(class_id.strip())] = class_name.strip()
                    else:
                        messagebox.showerror("Class file configuration error", "Invalid class mapping format. Each line should be in 'id:class_name' format.")
                        return
                    
                self.general_data["class_mapping"] = class_mapping
                    
                if self.general_data["label_format"].get() == LabelFormat.MASK.value:
                    self.update_color_map()
                messagebox.showinfo("Success", "Class mapping file loaded successfully.")
        except Exception as e:
                messagebox.showerror("Class file configuration error", f"Failed to load file: {e}")

    def delete_file(self):
        self.general_data["class_map_file"].set("")
        self.general_data["class_mapping"] = {}

    def select_folder(self, key):
        folder = filedialog.askdirectory(title=f"Select {key.replace('_', ' ')}")
        if folder:
            self.general_data[key].set(folder)
            

    def on_format_change(self):
        label_format = self.general_data["label_format"].get()
        
        show_image_related = bool(label_format)  
        show_class_mapping = label_format != "" and label_format != "json"  
        show_label_file = label_format == LabelFormat.JSON.value  
        show_color_dict = label_format == LabelFormat.MASK.value

        for widget in [self.color_map_label, self.add_color_bt, self.transformation_color_dict_text, 
                    self.image_path_label, self.entry_images, self.img_path_bt,
                    self.label_path_label, self.label_entry, self.label_path_bt, self.label_file_bt,
                    self.output_path_label, self.output_entry, self.output_path_bt,
                    self.verbose_label, self.verbose_checkbutton,
                    self.binary_label, self.binary_checkbutton,
                    self.background_label, self.background_entry]:
            widget.grid_remove()
        
        if show_image_related:
            for widget in [self.image_path_label, self.entry_images, self.img_path_bt,
                        self.output_path_label, self.output_entry, self.output_path_bt,
                        self.verbose_label, self.verbose_checkbutton,
                        self.binary_label, self.binary_checkbutton,
                        self.background_label, self.background_entry]:
                widget.grid()

        if show_class_mapping:
            for widget in [self.class_map_label, self.class_mapping, self.class_map_bt]:
                widget.grid()
        else:
            for widget in [self.class_map_label, self.class_mapping, self.class_map_bt]:
                widget.grid_remove()
        
        if label_format:
            if show_color_dict:
                self.color_map_label.grid()
                self.add_color_bt.grid()
                self.transformation_color_dict_text.grid()
            else:
                self.color_map_label.grid_remove()
                self.add_color_bt.grid_remove()
                self.transformation_color_dict_text.grid_remove()

            if show_label_file:
                self.label_path_bt.grid_remove()
                self.label_file_bt.grid()
            else:
                self.label_file_bt.grid_remove()
                self.label_path_bt.grid()
        
        if show_image_related:
            for widget in [self.label_path_label, self.label_entry]:
                widget.grid()

    def validate_form(self):
        
        class_map_path = self.general_data["class_map_file"].get().strip()
        if class_map_path != "" and not self.general_data["class_mapping"]:  
            result_warning = messagebox.askokcancel("Warning", "Class map file path defined but not loaded. If you continue, the map file will not be used.")

            if not result_warning:
                return False
            
            self.general_data["class_map_file"] = tk.StringVar(value="")

        label_format = self.general_data["label_format"].get().strip()

        if label_format == "":
            tk.messagebox.showerror("General configuration error", "You must choose a label format before proceeding.")
            return False
        image_path = self.general_data["image_path"].get().strip()
        
        if image_path == "":
            tk.messagebox.showerror("General configuration error", "You must select an image path before proceeding.")
            return False

        images = [f for f in os.listdir(image_path) if os.path.isfile(os.path.join(image_path, f))]
        if len(images) == 0:
            tk.messagebox.showerror("General configuration error", "The selected image folder does not contain images.")
            return False

        non_images = [f for f in images if os.path.isfile(os.path.join(image_path, f)) and not f.lower().endswith(tuple(list(ConfigHandler.VALID_IMAGE_EXTENSIONS)))]

        if non_images:
            if len(non_images) <10:   
                messagebox.showerror("General configuration error", f"The folder contains non-images files:\n{', '.join(non_images)}")
            else:
                messagebox.showerror("General configuration error", f"The folder contains non-images files (showing first 10 invalid files):\n{', '.join(non_images[:10])}")

            return False
        
        label_path = self.general_data["label_path"].get().strip()

        if label_path == "":
            tk.messagebox.showerror("General configuration error", "You must select a label path before proceeding.")
            return False
        
        label_ext = f".{label_format}"
        if label_format == LabelFormat.MASK.value:
            label_ext = ConfigHandler.VALID_IMAGE_EXTENSIONS

        if label_format != "json":

            label_files = [f for f in os.listdir(label_path) if os.path.isfile(os.path.join(label_path, f))]
            if len(label_files) == 0:
                tk.messagebox.showerror("General configuration error", "The selected label folder does not contain images.")
                return False

            non_valid_labels = [f for f in label_files if os.path.isfile(os.path.join(label_path, f)) and not f.lower().endswith(tuple(label_ext))]

            if non_valid_labels:
                if len(non_valid_labels) <10:   
                    messagebox.showerror("General configuration error", f"The folder contains labels with invalid extensions:\n{', '.join(non_valid_labels)}")
                else:
                    messagebox.showerror("General configuration error", f"The folder contains labels with invalid extensions (showing first 10 invalid files):\n{', '.join(non_valid_labels[:10])}...")

                return False
            
            if len(label_files) != len(images):
                img_names = {os.path.splitext(image)[0] for image in images}
                label_names = {os.path.splitext(label)[0] for label in label_files}

                images_not_in_labels = img_names - label_names

                images_unmatched = [f for f in images if os.path.splitext(f)[0] in images_not_in_labels]

                move_images = messagebox.askyesno("Missing label files", f"The following images don't have label files in the label directory:\n{', '.join(images_unmatched)} \n"
                                    "All images need to have their respective labels for proper training.\nDo you wish to move these images automatically?")

                if not move_images:
                    return False
                
                save_path = os.path.join(os.path.dirname(image_path), "ignored_images")

                os.makedirs(save_path,exist_ok=True)

                for img in images_unmatched:
                    src = os.path.join(image_path, img)
                    dst = os.path.join(save_path, img)
                    shutil.move(src, dst)
                
                messagebox.showinfo("Missing label files", f"Images without labels have been moved to {save_path}")

        output_path = self.general_data["output_path"].get().strip()

        if output_path == "":
            tk.messagebox.showerror("General configuration error", "You must select an output path before proceeding.")
            return False
        
        if self.general_data["binary"].get() and self.general_data["threshold"].get().strip() == "":
            tk.messagebox.showerror("General configuration error", "You must select a threshold when the labels are binary.")
            return False
        
        colors = self.general_data["color_dict"]
        
        if self.general_data["color_dict"] and not all(color == "" for color in colors.values()):
        
            for class_id, color in colors.items():
                if color == "":
                    tk.messagebox.showerror("General configuration error", f"Missing color for {class_id}. All classes must have a color.")
                    return False
        
        return True

    def toggle_load_button(self):
        if self.general_data["class_map_file"].get().strip() != "":
            self.clear_classmap_bt.grid()
        else:
            self.clear_classmap_bt.grid_remove()

    def next(self):
        
        validation_ok = self.validate_form()
        if validation_ok:
            self.config_data["general_data"].update(self.general_data)
            self.controller.show_frame("AnalysisConfigFrame")



    



from segdan.models.semanticsegmentationmodel import SemanticSegmentationModel
from segdan.models.hfstransformermodel import HFTransformerModel
from segdan.models.smpmodel import SMPModel

class SemanticSegmentationModelFactory:

    def __init__(self):
        self.framework_map = {
            "smp": (SMPModel, {"in_channels", "t_max"}),
            "huggingface": (HFTransformerModel, {"batch_size"})
        }

    def get_model(self, framework: str, args: dict) -> SemanticSegmentationModel:
        try:
            model_class, required_params = self.converter_map[framework]
        except KeyError:
            raise ValueError(f"Unsupported semantic segmentation framework: {framework}")
        
        missing = required_params - args.keys()
        if missing:
            raise ValueError(f"Missing required context params: {missing} for model {model_class.__name__}")
        
        init_args = {key: args[key] for key in required_params if key in args}
        return model_class(**init_args)


import tkinter as tk

class FormUtils():

    @staticmethod
    def toggle_widget(widget, variable):
        if variable.get():  
            widget.grid()  
        else:
            widget.grid_remove()  

    @staticmethod
    def toggle_label_entry(event, label, entry, comment, row, column):

        if event.get():
            label.grid(row=row, column=column, padx=5)
            entry.grid(row=row+1, column=column, padx=5)
            if comment is not None:
                comment.grid(row=row+2, column=column, padx=5)
        else:
            label.grid_forget()
            entry.grid_forget()
            if comment is not None:
                comment.grid_forget()

    @staticmethod
    def convert_value(value):
            if isinstance(value, tk.BooleanVar):
                return value.get()
            if isinstance(value, tk.StringVar):
                val = value.get()
                try:
                    return int(val)
                except ValueError:
                    pass
                try:
                    return float(val)
                except ValueError:
                    return val.lower()
            else:
                return value

    @staticmethod
    def save_config(data):
        config_data = {}

        for key, value in data.items():
            if isinstance(value, dict):
                config_data[key] = FormUtils.save_config(value)
            else:
                config_data[key] = FormUtils.convert_value(value) if isinstance(value, (tk.StringVar, tk.IntVar, tk.DoubleVar, tk.BooleanVar)) else value


        return config_data

    @staticmethod
    def center_window(child, parent):
        parent.update_idletasks()  

        width = child.top.winfo_width()  
        height = child.top.winfo_height()  

        if width == 1 or height == 1:  
            width = child.top.winfo_reqwidth()
            height = child.top.winfo_reqheight()

        x = parent.winfo_rootx() + (parent.winfo_width() // 2) - (width // 2)
        y = parent.winfo_rooty() + (parent.winfo_height() // 2) - (height // 2)

        child.top.geometry(f'+{x}+{y}')



import os
from typing import Optional
import cv2
import numpy as np
from PIL import Image
import pandas as pd
from tqdm import tqdm

from segdan.exceptions.exceptions import ExtensionNotFoundException
from segdan.extensions.extensions import LabelExtensions

class ImageLabelUtils:

    @staticmethod
    def label_to_image(label_path: str, img_path:str , img_ext:str ) -> str: 

        if not os.path.exists(label_path):
            raise FileNotFoundError(f"Directory {label_path} does not exist.")

        if not os.path.exists(img_path):
            raise FileNotFoundError(f"Directory {img_path} does not exist.")

        label_name, _ = os.path.splitext(os.path.basename(label_path))

        image = os.path.join(img_path, f"{label_name}{img_ext}")

        return image
    
    @staticmethod
    def image_to_label(img_path: str, label_path:str , lbl_ext:str ) -> str: 

        if not os.path.exists(img_path):
            raise FileNotFoundError(f"Directory {img_path} does not exist.")

        if not os.path.exists(label_path):
            raise FileNotFoundError(f"Directory {label_path} does not exist.")
        
        image_name, _ = os.path.splitext(os.path.basename(img_path))

        label = os.path.join(label_path, f"{image_name}{lbl_ext}")

        return label
    
    @staticmethod
    def save_multilabel_mask(mask, file_name, output_dir: str):
        mask_filename = os.path.splitext(file_name)[0] + ".png"
        mask_filepath = os.path.join(output_dir, mask_filename)
        cv2.imwrite(mask_filepath, mask)

    @staticmethod
    def get_labels_dir(dataset_path: str):
        try:
            labels_dir = os.path.join(dataset_path, "labels")
            if os.path.isdir(labels_dir):
                return labels_dir
            return None
        except Exception as e:
            return None

    @staticmethod
    def check_label_extensions(label_dir, verbose=False, logger=None):
        """
        Checks the extensions of label files in the label directory to ensure consistency.

        Args:
            verbose (bool): If True, logs the process and any issues found.

        Returns:
            str: The extension of the label files if consistent.

        Raises:
            ValueError: If multiple extensions are found in the label directory.
            ExtensionNotFoundException: If the extension is not recognized by the system.
        """
        
        if verbose:
            logger.info(f"Checking label extensions from path: {label_dir}...")

        labels_ext = {os.path.splitext(file)[1] for file in os.listdir(label_dir)}

        if len(labels_ext) == 1:
            ext = labels_ext.pop()  
            try:
                enum_ext = LabelExtensions.extensionToEnum(ext)

                if enum_ext:
                    if verbose:
                        logger.info(f"All labels are in {enum_ext.name} format.")

                    return LabelExtensions.enumToExtension(enum_ext)
                
            except ExtensionNotFoundException as e:
                print(f"All labels are in unknown {ext} format.")
                raise e
        else:
            raise ValueError(f"The directory contains multiple extensions for labels: {labels_ext}.")
        
    @staticmethod
    def all_images_are_color(directory):
        for filename in os.listdir(directory):
            image_path = os.path.join(directory, filename)
            if not os.path.isfile(image_path):
                continue

            image = cv2.imread(image_path, cv2.IMREAD_UNCHANGED)
            if image is None:
                continue  

            if len(image.shape) == 2:
                return False

            if image.shape[2] >= 3:
                b, g, r = image[..., 0], image[..., 1], image[..., 2]
                if np.array_equal(b, g) and np.array_equal(b, r):
                    return False  

        return True
    
    @staticmethod
    def get_classes_from_csv(csv_file, background: Optional[int]=None):
        df = pd.read_csv(csv_file, delimiter=";")

        class_names_dict = {i: str(class_name) for i, class_name in enumerate(df["Class name"].values)}

        if background:
            class_names_dict[background] = "background"

        return class_names_dict.values()

    @staticmethod
    def count_num_classes_png(label_dir):
        all_classes = set()

        for filename in tqdm(os.listdir(label_dir), desc="Counting number of classes from multiclass labels..."):
            if filename.endswith('.png'):
                path = os.path.join(label_dir, filename)
                mask = np.array(Image.open(path))
                unique_classes = np.unique(mask)
                all_classes.update(unique_classes)

        return len(all_classes), all_classes







import torch

from segdan.metrics.segmentationmetrics import accuracy, iou_score, dice_score, precision, recall, f1_score
from segdan.metrics import custom_metric

metric_functions = {
    "accuracy": accuracy,
    "iou": iou_score,
    "dice": dice_score,
    "precision": precision,
    "recall": recall,
    "f1": f1_score,
}

def compute_metrics(results, metrics, stage="train"):
    tp = torch.cat([x["tp"] for x in results])
    fp = torch.cat([x["fp"] for x in results])
    fn = torch.cat([x["fn"] for x in results])
    tn = torch.cat([x["tn"] for x in results])

    results = {}

    for metric in metrics:
        metric_fn = metric_functions[metric]

        score_global = custom_metric(tp, fp, fn, tn, metric_fn, reduction="micro")
        results[f"{metric}_{stage}"] = score_global.item() if torch.is_tensor(score_global) else score_global

        score_none = custom_metric(tp, fp, fn, tn, metric_fn, reduction="none")
        score_per_class = score_none.mean(dim=0)  

        for class_idx, class_score in enumerate(score_per_class):
            results[f"{metric}_{stage}_class_{class_idx}"] = class_score.item()

    return results